package main

import (
	"context"
	"database/sql"
	"flag"
	"fmt"
	"log"
	"os"
	"os/signal"
	"runtime"
	"sync"
	"sync/atomic"
	"syscall"
	"time"

	_ "github.com/lib/pq"
	"github.com/tidwall/btree"

	_ "github.com/lib/pq"

	"github.com/dustin/go-humanize"
	"github.com/minio/minio-go/v7"
	"github.com/minio/minio-go/v7/pkg/credentials"
)

type cleanner struct {
	context.Context
	bucket   string
	mc       *minio.Client
	dryRun   bool
	fileIds  *btree.Set[string]
	objects  map[string]*xobject
	lock     sync.Locker
	planned  stats
	removed  stats
	analysed stats
}

type stats struct {
	objects  int64
	versions int64
	space    int64
}

type xobject struct {
	sync.Locker
	key      string
	garbage  bool
	versions []minio.ObjectInfo
}

func main() {
	startTime := time.Now()

	fmt.Println("fetching in-database file ids")
	fileIds, err := getFileIds()
	CheckError(err)

	ctx := context.Background()
	ctx, stop := signal.NotifyContext(ctx, syscall.SIGINT, syscall.SIGTERM)
	defer stop()

	minioClient, err := minio.New(os.Getenv("S3_ENDPOINT"), &minio.Options{
		Creds:  credentials.NewStaticV4(os.Getenv("S3_ACCESS_KEY"), os.Getenv("S3_SECRET_KEY"), ""),
		Secure: true,
	})
	if err != nil {
		log.Fatalln(err)
	}

	c := cleanner{
		Context: ctx,
		bucket:  os.Getenv("S3_BUCKET"),
		mc:      minioClient,
		fileIds: fileIds,
		objects: map[string]*xobject{},
		lock:    &sync.Mutex{},
	}

	flag.BoolVar(&c.dryRun, "dry-run", false, "")
	flag.Parse()
	if c.dryRun {
		log.Printf("dry run enabled, nothing will be removed\n")
	}

	log.Printf("cleanning %s bucket\n", c.bucket)

	all := c.getAllObjects()
	changes := c.processObjects(all)
	<-c.notifyProgress(changes)
	c.removeXObjects()

	log.Printf("took %v\n", time.Since(startTime))
	if !c.dryRun {
		log.Printf("removed %d objects & %d versions\n", c.removed.objects, c.removed.versions)
		log.Printf("reclaimed %s\n", humanize.Bytes(uint64(c.removed.space)))
	}
}

func CheckError(err error) {
	if err != nil {
		log.Fatalln(err)
	}
}

func getFileIds() (*btree.Set[string], error) {
	var fileIds btree.Set[string]
	db, err := sql.Open("postgres", fmt.Sprintf("host=%s port=%s user=%s password=%s dbname=%s sslmode=require",
		os.Getenv("PG_HOST"),
		os.Getenv("PG_PORT"),
		os.Getenv("PG_USER"),
		os.Getenv("PG_PASS"),
		os.Getenv("PG_DB")))
	if err != nil {
		return nil, err
	}
	defer db.Close()

	query := "SELECT fileid FROM filecache;"
	table := os.Getenv("PG_TABLE")
	if table == "oc_filecache" {
		query = "SELECT fileid FROM oc_filecache;"
	}
	rows, err := db.Query(query)
	if err != nil {
		return nil, err
	}
	for rows.Next() {
		var id int
		err = rows.Scan(&id)
		if err != nil {
			return nil, err
		}
		fileId := fmt.Sprintf("urn:oid:%d", id)
		fileIds.Insert(fileId)
	}
	return &fileIds, nil
}

func (c *cleanner) xobject(key string) *xobject {
	c.lock.Lock()
	xo := c.objects[key]
	if xo == nil {
		xo = &xobject{
			Locker:   &sync.Mutex{},
			key:      key,
			versions: []minio.ObjectInfo{},
		}
		c.objects[key] = xo
	}
	c.lock.Unlock()
	return xo
}

func (c *cleanner) getAllObjects() <-chan minio.ObjectInfo {
	return c.mc.ListObjects(c, c.bucket, minio.ListObjectsOptions{
		Recursive:    true,
		WithVersions: true,
		WithMetadata: true,
	})
}

func (c *cleanner) processObjects(objects <-chan minio.ObjectInfo) <-chan struct{} {
	changes := make(chan struct{})
	go func() {
		wg := sync.WaitGroup{}
		for object := range objects {
			wg.Add(1)
			go func(object minio.ObjectInfo) {
				xo := c.xobject(object.Key)
				xo.Lock()
				xo.versions = append(xo.versions, object)
				if !xo.garbage && (object.IsDeleteMarker && object.IsLatest || !c.fileIds.Contains(object.Key)) {
					xo.garbage = true
					atomic.AddInt64(&c.planned.objects, 1)
				}
				atomic.AddInt64(&c.analysed.versions, 1)
				atomic.AddInt64(&c.analysed.space, object.Size)
				if xo.garbage {
					atomic.AddInt64(&c.planned.versions, 1)
					atomic.AddInt64(&c.planned.space, object.Size)
					changes <- struct{}{}
				}
				xo.Unlock()
				wg.Done()
			}(object)
		}
		wg.Wait()
		close(changes)
	}()
	return changes
}

func (c *cleanner) notifyProgress(changes <-chan struct{}) <-chan struct{} {
	done := make(chan struct{})
	start := time.Now()
	go func() {
		for range changes {
			v := c.planned.versions * 100 / c.analysed.versions
			var s int64
			if c.analysed.space > 0 && c.planned.space > 0 {
				s = c.planned.space * 100 / c.analysed.space
			}
			vs := int64(0)
			diff := int64(time.Since(start).Seconds())
			if diff > 0 {
				vs = c.analysed.versions / diff
			}
			log.Printf("[will remove] %d objects : %d versions (%d%%) and %s (%d%%) | %d version/s\n", c.planned.objects, c.planned.versions, v, humanize.Bytes(uint64(c.planned.space)), s, vs)
		}
		done <- struct{}{}
		close(done)
	}()
	return done
}

func (c *cleanner) removeXObjects() {
	xobjects := make(chan *xobject)
	wg := sync.WaitGroup{}
	for i := 0; i < runtime.NumCPU(); i++ {
		wg.Add(1)
		log.Printf("[starting] remove worker %d\n", i)
		go c.removeWorker(&wg, xobjects)
	}
	go func() {
		for _, xo := range c.objects {
			select {
			case <-c.Done():
				break
			default:
				if xo.garbage {
					if c.fileIds.Contains(xo.key) {
						log.Printf("[error] object %s is still referenced by the database\n", xo.key)
						continue
					}
					xobjects <- xo
				}
			}
		}
		close(xobjects)
	}()
	wg.Wait()
}

func (c *cleanner) removeWorker(wg *sync.WaitGroup, xobjects chan *xobject) {
	for xo := range xobjects {
		if !c.dryRun {
			c.removeXObject(xo)
		}
	}
	wg.Done()
}

func (c *cleanner) removeXObject(xobject *xobject) {
	versions := len(xobject.versions)
	log.Printf("[removing] %s %d versions\n", xobject.key, versions)
	for _, v := range xobject.versions {
		err := c.mc.RemoveObject(c, c.bucket, v.Key, minio.RemoveObjectOptions{VersionID: v.VersionID})
		if err != nil {
			log.Printf("failed to remove %s version %s %s\n", v.Key, v.VersionID, err)
		}
		atomic.AddInt64(&c.removed.versions, int64(versions))
		atomic.AddInt64(&c.removed.space, v.Size)
	}
	atomic.AddInt64(&c.removed.objects, 1)
}
