package cmd

import (
	"context"
	"errors"
	"fmt"
	"log"
	"os/signal"
	"sync"
	"syscall"
	"time"

	"github.com/spf13/cobra"
	"libre.sh/tools/pkg/postgres"

	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/fields"
	"k8s.io/apimachinery/pkg/runtime/schema"
	"k8s.io/cli-runtime/pkg/genericclioptions"
	"k8s.io/client-go/dynamic"
	"k8s.io/client-go/rest"
	"k8s.io/kubectl/pkg/scheme"
)

type PosgresOptions struct {
	configFlags *genericclioptions.ConfigFlags
	genericclioptions.IOStreams
	ctx           context.Context
	allNamespaces bool
	errors        chan error
}

func NewPostgresOptions(streams genericclioptions.IOStreams) *PosgresOptions {
	ctx, _ := signal.NotifyContext(context.Background(), syscall.SIGINT, syscall.SIGTERM)
	o := &PosgresOptions{
		configFlags: genericclioptions.NewConfigFlags(true),
		IOStreams:   streams,
		ctx:         ctx,
		errors:      make(chan error),
	}
	return o
}

func NewCmdPostgres(streams genericclioptions.IOStreams) *cobra.Command {
	o := NewPostgresOptions(streams)

	cmd := &cobra.Command{
		Use:          "postgres [command] [flags]",
		SilenceUsage: true,
	}

	o.configFlags.AddFlags(cmd.PersistentFlags())
	cmd.PersistentFlags().BoolVarP(&o.allNamespaces, "all-namespaces", "A", false, "")

	cmd.AddCommand(
		NewCmdPostgresBackup(o),
		NewCmdPostgresSwitchover(o),
		NewCmdPostgresDestroyReplicas(o),
	)

	return cmd
}

func NewCmdPostgresBackup(o *PosgresOptions) *cobra.Command {
	var logical bool
	var skipNoCronJob bool
	cmd := &cobra.Command{
		Use:          "backup [name]",
		SilenceUsage: true,
		RunE: func(cmd *cobra.Command, args []string) error {
			var name string
			if len(args) > 0 {
				name = args[0]
			}
			return o.forEachPostgres(name, "not all backups were successful", func(pg *postgres.Postgres) error {
				log.Printf("Start backup for %s\n", pg.GetNamespaceName())
				if logical {
					if err := pg.LogicalBackup(); err != nil {
						if !skipNoCronJob || !errors.Is(err, postgres.ErrNoCronJob) {
							return err
						}
					} else {
						log.Printf("Wrote backup for %s\n", pg.GetNamespaceName())
					}
				} else {
					msg, err := pg.Backup()
					if err != nil {
						return err
					}
					log.Printf("Wrote backup %s for %s\n", msg, pg.GetNamespaceName())
				}
				return nil
			})
		},
	}
	cmd.Flags().BoolVar(&logical, "logical", false, "create a logical backup (or dump) instead")
	cmd.Flags().BoolVar(&skipNoCronJob, "skip-no-cronjob", false, "don't throw an error if no cronjob is found")
	return cmd
}

func NewCmdPostgresSwitchover(o *PosgresOptions) *cobra.Command {
	cmd := &cobra.Command{
		Use:          "switchover [name]",
		SilenceUsage: true,
		RunE: func(cmd *cobra.Command, args []string) error {
			var name string
			if len(args) > 0 {
				name = args[0]
			}
			return o.forEachPostgres(name, "not all swithovers were successful", func(pg *postgres.Postgres) error {
				log.Printf("Start switchover for %s\n", pg.GetNamespaceName())
				if err := pg.IsHealthy(); err != nil {
					return err
				}
				if err := pg.Switchover(); err != nil {
					return err
				}
				if err := pg.WaitHealthy(time.Second * 30); err != nil {
					return err
				}
				log.Printf("Succeeded switchover for %s\n", pg.GetNamespaceName())
				return nil
			})
		},
	}
	return cmd
}

func NewCmdPostgresDestroyReplicas(o *PosgresOptions) *cobra.Command {
	cmd := &cobra.Command{
		Use:          "destroy-replicas [name]",
		SilenceUsage: true,
		RunE: func(cmd *cobra.Command, args []string) error {
			var name string
			if len(args) > 0 {
				name = args[0]
			}
			return o.forEachPostgres(name, "failed to destroy all replicas", func(pg *postgres.Postgres) error {
				log.Printf("Start destroying replicas for %s\n", pg.GetNamespaceName())
				if err := pg.IsHealthy(); err != nil {
					return err
				}
				if err := pg.DestroyReplicas(); err != nil {
					return err
				}
				if err := pg.WaitHealthy(time.Minute * 5); err != nil {
					return err
				}
				log.Printf("Succeeded destroyed replicas for %s\n", pg.GetNamespaceName())
				return nil
			})
		},
	}
	return cmd
}

func (o *PosgresOptions) getNamespace() (string, error) {
	if o.allNamespaces {
		return metav1.NamespaceAll, nil
	}
	ns, _, err := o.configFlags.ToRawKubeConfigLoader().Namespace()
	return ns, err
}

func (o *PosgresOptions) getPostgresList(name string) ([]unstructured.Unstructured, error) {
	restConf, err := o.configFlags.ToRESTConfig()
	if err != nil {
		return nil, err
	}
	client, err := dynamic.NewForConfig(restConf)
	if err != nil {
		return nil, err
	}
	ns, err := o.getNamespace()
	if err != nil {
		return nil, err
	}
	selectors := []fields.Selector{}
	if !o.allNamespaces {
		selectors = append(selectors, fields.OneTermEqualSelector("metadata.namespace", ns))
	}
	if name != "" {
		selectors = append(selectors, fields.OneTermEqualSelector("metadata.name", name))
	}
	opts := metav1.ListOptions{}
	if len(selectors) > 0 {
		opts.FieldSelector = fields.AndSelectors(selectors...).String()
	}
	list, err := client.Resource(postgres.GroupVersionResource).List(o.ctx, opts)
	if err != nil {
		return nil, err
	}
	if len(list.Items) == 0 {
		return nil, errors.New("no resource found")
	}
	return list.Items, nil
}

func (o *PosgresOptions) forEachPostgres(name string, errorMsg string, f func(*postgres.Postgres) error) error {
	list, err := o.getPostgresList(name)
	if err != nil {
		return err
	}
	restConf, err := o.configFlags.ToRESTConfig()
	if err != nil {
		return err
	}
	if err := setKubernetesDefaults(restConf); err != nil {
		return err
	}
	errs := []error{}
	errLock := sync.Mutex{}
	wg := sync.WaitGroup{}
	for _, upg := range list {
		wg.Add(1)
		pg := postgres.For(o.ctx, restConf, upg)
		go func() {
			defer wg.Done()
			if err := f(pg); err != nil {
				errLock.Lock()
				errs = append(errs, err)
				errLock.Unlock()
			}
		}()
	}
	wg.Wait()
	if len(errs) > 0 {
		for _, err := range errs {
			log.Println(err)
		}
		return fmt.Errorf(errorMsg)
	}
	return nil
}

func setKubernetesDefaults(config *rest.Config) error {
	config.GroupVersion = &schema.GroupVersion{Group: "", Version: "v1"}
	if config.APIPath == "" {
		config.APIPath = "/api"
	}
	if config.NegotiatedSerializer == nil {
		config.NegotiatedSerializer = scheme.Codecs.WithoutConversion()
	}
	return rest.SetKubernetesDefaults(config)
}
