package postgres

import (
	"bytes"
	"fmt"
	"io"
	"regexp"
	"time"

	batchv1 "k8s.io/api/batch/v1"
	batchv1beta1 "k8s.io/api/batch/v1beta1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/fields"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/client-go/kubernetes"
)

var backupRegex = regexp.MustCompile(`(?m)Wrote backup with name (.+)$`)

func (pg *Postgres) Backup() (string, error) {
	master, err := pg.GetMasterPod()
	if err != nil {
		return "", err
	}
	var out bytes.Buffer
	err = pg.Execute(master, []string{"su", "-", "postgres", "-c", `envdir "/run/etc/wal-e.d/env" /scripts/postgres_backup.sh "/home/postgres/pgdata/pgroot/data"`}, &out)
	if err != nil {
		return "", err
	}
	logs, err := io.ReadAll(&out)
	if err != nil {
		return "", err
	}
	matches := backupRegex.FindSubmatch(logs)
	if matches == nil {
		return "", fmt.Errorf("failed to backup %s\n%s", pg.resource.GetName(), logs)
	}
	return string(matches[1]), nil
}

func (pg *Postgres) LogicalBackup() error {
	cronJob, err := pg.GetCronJob()
	if err != nil {
		return err
	}
	job := createJobFromCronJob(&cronJob)
	client, err := kubernetes.NewForConfig(pg.restConfig)
	if err != nil {
		return err
	}
	job, err = client.BatchV1().Jobs(pg.GetNamespace()).Create(pg.ctx, job, metav1.CreateOptions{})
	if err != nil {
		return err
	}
	watch, err := client.BatchV1().Jobs(pg.GetNamespace()).Watch(pg.ctx, metav1.ListOptions{
		FieldSelector: fields.OneTermEqualSelector("metadata.name", job.Name).String(),
	})
	if err != nil {
		return err
	}
	for range watch.ResultChan() {
		job, err = client.BatchV1().Jobs(pg.GetNamespace()).Get(pg.ctx, job.Name, metav1.GetOptions{})
		if err != nil {
			return err
		}
		if job.Status.Failed > 0 {
			watch.Stop()
			return pg.WrapError(ErrCronJobFailed)
		}
		if job.Status.Succeeded > 0 {
			break
		}
	}
	return nil
}

func (pg *Postgres) GetCronJob() (batchv1beta1.CronJob, error) {
	client, err := kubernetes.NewForConfig(pg.restConfig)
	if err != nil {
		return batchv1beta1.CronJob{}, err
	}
	list, err := client.BatchV1beta1().CronJobs(pg.GetNamespace()).List(pg.ctx, metav1.ListOptions{
		LabelSelector: labels.FormatLabels(map[string]string{
			"cluster-name": pg.GetName(),
		}),
	})
	if err != nil {
		return batchv1beta1.CronJob{}, err
	}
	if len(list.Items) > 1 {
		return batchv1beta1.CronJob{}, pg.WrapError(ErrMultipleCronJobs)
	}
	if len(list.Items) == 0 {
		return batchv1beta1.CronJob{}, pg.WrapError(ErrNoCronJob)
	}
	return list.Items[0], nil
}

func createJobFromCronJob(cronJob *batchv1beta1.CronJob) *batchv1.Job {
	annotations := make(map[string]string)
	annotations["cronjob.kubernetes.io/instantiate"] = "manual"
	for k, v := range cronJob.Spec.JobTemplate.Annotations {
		annotations[k] = v
	}
	job := batchv1.Job{
		ObjectMeta: metav1.ObjectMeta{
			Name:        fmt.Sprintf("%s-%d", cronJob.GetName(), time.Now().Unix()),
			Namespace:   cronJob.Namespace,
			Annotations: annotations,
			Labels:      cronJob.Spec.JobTemplate.Labels,
			OwnerReferences: []metav1.OwnerReference{
				{
					APIVersion: batchv1.SchemeGroupVersion.String(),
					Kind:       "CronJob",
					Name:       cronJob.GetName(),
					UID:        cronJob.GetUID(),
				},
			},
		},
		Spec: cronJob.Spec.JobTemplate.Spec,
	}
	return &job
}
