package postgres

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"regexp"
	"time"

	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/runtime/schema"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/tools/remotecommand"
	"k8s.io/kubectl/pkg/scheme"
)

var GroupVersionResource = schema.GroupVersionResource{
	Group:    "acid.zalan.do",
	Version:  "v1",
	Resource: "postgresqls",
}

type Role string

var (
	MasterRole  Role = "master"
	ReplicaRole Role = "replica"
)

type Postgres struct {
	restConfig *rest.Config
	resource   unstructured.Unstructured
	ctx        context.Context
}

func For(ctx context.Context, restConfig *rest.Config, resource unstructured.Unstructured) *Postgres {
	return &Postgres{
		ctx:        ctx,
		restConfig: restConfig,
		resource:   resource,
	}
}

func (pg *Postgres) GetName() string {
	return pg.resource.GetName()
}

func (pg *Postgres) GetNamespace() string {
	return pg.resource.GetNamespace()
}

func (pg *Postgres) GetNamespaceName() string {
	return fmt.Sprintf("%s/%s", pg.GetNamespace(), pg.GetName())
}

func (pg *Postgres) GetPods(role Role) ([]corev1.Pod, error) {
	client, err := kubernetes.NewForConfig(pg.restConfig)
	if err != nil {
		return nil, err
	}
	name := pg.resource.GetName()
	ns := pg.resource.GetNamespace()
	list, err := client.CoreV1().Pods(ns).List(pg.ctx, metav1.ListOptions{
		LabelSelector: labels.FormatLabels(map[string]string{
			"cluster-name": name,
			"spilo-role":   string(role),
		}),
	})
	if err != nil {
		return nil, err
	}
	if len(list.Items) == 0 {
		return nil, fmt.Errorf("no %s pod found for cluster %s", role, name)
	}
	return list.Items, nil
}

func (pg *Postgres) GetMasterPod() (corev1.Pod, error) {
	pods, err := pg.GetPods(MasterRole)
	if err != nil {
		return corev1.Pod{}, err
	}
	if len(pods) > 1 {
		return corev1.Pod{}, fmt.Errorf("cluster %s has multiple master", pg.resource.GetName())
	}
	return pods[0], nil
}

func (pg *Postgres) Execute(pod corev1.Pod, command []string, stdout io.Writer) error {
	client, err := rest.RESTClientFor(pg.restConfig)
	if err != nil {
		return err
	}
	req := client.Post().
		Resource("pods").
		Name(pod.Name).
		Namespace(pod.Namespace).
		SubResource("exec").
		VersionedParams(&corev1.PodExecOptions{
			TypeMeta: metav1.TypeMeta{Kind: "Pod"},
			Command:  command,
			Stdout:   stdout != nil,
			Stderr:   stdout != nil,
			TTY:      false,
		}, scheme.ParameterCodec)
	executor, err := remotecommand.NewSPDYExecutor(pg.restConfig, "POST", req.URL())
	if err != nil {
		return err
	}
	return executor.StreamWithContext(pg.ctx, remotecommand.StreamOptions{
		Stdout: stdout,
		Stderr: stdout,
		Tty:    false,
	})
}

func (pg *Postgres) DeclaredInstances() int64 {
	spec := pg.resource.Object["spec"].(map[string]interface{})
	return spec["numberOfInstances"].(int64)
}

type patroniMember struct {
	State string `json:"State"`
	TL    int64  `json:"TL"`
	Lag   int64  `json:"Lag in MB"`
}

func (pg *Postgres) IsHealthy() error {
	master, err := pg.GetMasterPod()
	if err != nil {
		return err
	}
	var out bytes.Buffer
	err = pg.Execute(master, []string{"patronictl", "list", "-f", "json"}, &out)
	if err != nil {
		return err
	}
	logs := out.Bytes()
	var members []patroniMember
	if err = json.Unmarshal(logs, &members); err != nil {
		return pg.WrapError(ErrInvalidMemberList)
	}
	if int64(len(members)) != pg.DeclaredInstances() {
		return pg.WrapError(ErrInvalidMemberCount)
	}
	var tl int64 = -1
	for _, member := range members {
		if member.State != "running" {
			return pg.WrapError(ErrMemberNotRunning)
		}
		if tl == -1 {
			tl = member.TL
		} else if tl != member.TL {
			return pg.WrapError(ErrMembersOutOfSync)
		}
	}
	return nil
}

func (pg *Postgres) Switchover() error {
	master, err := pg.GetMasterPod()
	if err != nil {
		return err
	}
	var out bytes.Buffer
	err = pg.Execute(master, []string{"patronictl", "switchover", "--force"}, &out)
	if err != nil {
		return err
	}
	logs := out.Bytes()
	regex, err := regexp.Compile(`(?m)Successfully switched over to`)
	if err != nil {
		return err
	}
	if !regex.Match(logs) {
		return pg.WrapError(fmt.Errorf("%s\n%w", logs, ErrSwitchoverFailed))
	}
	return nil
}

func (pg *Postgres) WaitHealthy(timeout time.Duration) error {
	start := time.Now()
	for {
		if time.Since(start) > timeout {
			return errors.New("wait timed out")
		}
		if err := pg.IsHealthy(); err == nil {
			return nil
		}
		time.Sleep(time.Second)
	}
}
